@@ -59,6 +59,8 @@ static char sccsid[] __attribute__((unused)) =
 #include <setjmp.h>
 #include <stdarg.h>
 #include <stdio.h>
+#include <stdint.h>
+#include <inttypes.h>
 #include <stdlib.h>
 #include <string.h>
 #include <time.h>
@@ -76,6 +78,9 @@ static char sccsid[] __attribute__((unused)) =
 #include <sys/time.h>
 #include <sys/uio.h>
 #include <sys/wait.h>
+#ifdef __linux__
+#include <sys/sysinfo.h>
+#endif
 
 #include <arpa/inet.h>
 #include <arpa/nameser.h>
@@ -117,6 +122,8 @@ struct filed consfile;
 
 static int	  Debug;		/* debug flag */
 static int	  Foreground = 0;	/* don't fork - don't run in daemon mode */
+static time_t	  boot_time;		/* Offset for printsys() */
+static uint64_t	  sys_seqno = 0;	/* Last seen kernel log message */
 static int	  resolve = 1;		/* resolve hostname */
 static char	  LocalHostName[MAXHOSTNAMELEN + 1]; /* our hostname */
 static char	 *LocalDomain;			     /* our local domain name */
@@ -165,6 +172,7 @@ void        doflush(void *arg);
 void        debug_switch();
 void        die(int sig);
 static void signal_init(void);
+static void boot_time_init(void);
 static void init(void);
 static int  strtobytes(char *arg);
 static int  cfparse(FILE *fp, struct files *newf);
@@ -189,6 +197,54 @@ static int addpeer(struct peer *pe0)
 	return 0;
 }
 
+static void sys_seqno_load(void)
+{
+	char buf[32], *str;
+	FILE *fp;
+
+	fp = fopen(_PATH_CACHE, "r");
+	if (!fp)
+		return;
+
+	while ((str = fgets(buf, sizeof(buf), fp))) {
+		uint64_t val;
+		char *end;
+
+		if (str[strlen(str) - 1] == '\n')
+			str[strlen(str) - 1] = 0;
+
+		errno = 0;
+		val = strtoull(str, &end, 10);
+		if (val == 0 && end == str)
+			break;	/* str was not a number */
+		else if (val == ULLONG_MAX && errno)
+			break; /* the value of str does not fit in unsigned long long */
+		else if (*end)
+			break; /* str began with a number but has junk left over at the end */
+
+		sys_seqno = val;
+	}
+	fclose(fp);
+}
+
+static void sys_seqno_save(void)
+{
+	static uint64_t prev = 0;
+	FILE *fp;
+
+	if (prev == sys_seqno)
+		return;		/* no changes since last save */
+
+	fp = fopen(_PATH_CACHE, "w");
+	if (!fp)
+		return;		/* best effort, ignore any errors */
+
+	fprintf(fp, "%" PRIu64 "\n", sys_seqno);
+	fclose(fp);
+
+	prev = sys_seqno;
+}
+
 int usage(int code)
 {
 	printf("Usage:\n"
@@ -206,7 +262,7 @@ int usage(int code)
 	       "                                    name or a port number\n"
 	       "              domainname[:port]     Accept messages where the reverse address\n"
 	       "                                    lookup yields 'domainname' for the sender\n"
-	       "                                    adress.  'domainname' may contain special\n"
+	       "                                    address.  'domainname' may contain special\n"
 	       "                                    shell-style pattern characters like '*'\n"
 	       "\n"
 	       "  -b NAME   Bind to a specific address and/or port. Multiple -b options may be\n"
@@ -232,7 +288,7 @@ int usage(int code)
 	       "  -s        Operate in secure mode, do not log messages from remote machines.\n"
 	       "            If specified twice, no socket at all will be opened, which also\n"
 	       "            disables support for logging to remote machines.\n"
-	       "  -T        Use local time and date for messages recived from remote hosts\n"
+	       "  -T        Use local time and date for messages received from remote hosts\n"
 	       "  -?        Show this help text\n"
 	       "  -v        Show program version and exit\n"
 	       "\n"
@@ -378,16 +434,25 @@ int main(int argc, char *argv[])
 		setlinebuf(stdout);
 	}
 
-	/* Attempt to open kernel log pipe */
-	if (opensys(_PATH_KLOG))
-		warn("Kernel logging disabled, failed opening %s", _PATH_KLOG);
-	else
+	/*
+	 * Attempt to open kernel log pipe.  On Linux we prefer
+	 * /dev/kmsg and fall back to _PROC_KLOG, which on GLIBC
+	 * systems is /proc/kmsg, and /dev/klog on *BSD.
+	 */
+	sys_seqno_load();
+	if (opensys("/dev/kmsg")) {
+		if (opensys(_PATH_KLOG))
+			warn("Kernel logging disabled, failed opening %s", _PATH_KLOG);
+		else
+			kern_console_off();
+	} else
 		kern_console_off();
 
 	consfile.f_type = F_CONSOLE;
 	strlcpy(consfile.f_un.f_fname, ctty, sizeof(consfile.f_un.f_fname));
 
 	logit("Starting.\n");
+	boot_time_init();
 	signal_init();
 	init();
 
@@ -439,6 +504,8 @@ int main(int argc, char *argv[])
 
 		if (rc < 0 && errno != EINTR)
 			ERR("select()");
+
+		sys_seqno_save();
 	}
 }
 
@@ -454,17 +521,29 @@ static void kernel_cb(int fd, void *arg)
 	for (;;) {
 		i = read(fd, line + len, MAXLINE - 1 - len);
 		if (i > 0) {
-			line[i + len] = '\0';
+			line[i + len] = 0;
 		} else {
-			if (i < 0 && errno != EINTR && errno != EAGAIN) {
-				ERR("klog read()");
-				socket_close(fd);
+			if (i < 0) {
+				switch (errno) {
+				case EPIPE: /* linux, log buffer overrun */
+					ERRX("Kernel log buffer filling up too quick, "
+					     "or too small log buffer, "
+					     "adjust kernel CONFIG_LOG_BUF_SHIFT");
+				case EINTR:
+				case EAGAIN:
+					break;
+
+				default:
+					ERR("klog read()");
+					socket_close(fd);
+					break;
+				}
 			}
 			break;
 		}
 
 		for (p = line; (q = strchr(p, '\n')) != NULL; p = q + 1) {
-			*q = '\0';
+			*q = 0;
 			printsys(p);
 		}
 		len = strlen(p);
@@ -475,8 +554,11 @@ static void kernel_cb(int fd, void *arg)
 		if (len > 0)
 			memmove(line, p, len + 1);
 	}
-	if (len > 0)
+
+	if (len > 0) {
+		line[len] = 0;
 		printsys(line);
+	}
 }
 
 static int opensys(const char *file)
@@ -507,6 +589,7 @@ static void unix_cb(int sd, void *arg)
 			ERR("UNIX recv()");
 		return;
 	}
+	msg[msglen] = 0;
 
 	parsemsg(LocalHostName, msg);
 }
@@ -577,6 +660,7 @@ static void inet_cb(int sd, void *arg)
 			ERR("INET recvfrom()");
 		return;
 	}
+	msg[len] = 0;
 
 	hname = cvthname((struct sockaddr *)&ss, sslen);
 	unmapped(sa);
@@ -806,7 +890,7 @@ parsemsg_rfc5424(const char *from, int pri, char *msg)
 	const char *start;						\
 									\
 	start = msg;							\
-	while (*msg >= '!' && *msg <= '~' && *msg != '=' &&		\
+	while (*msg && *msg >= '!' && *msg <= '~' && *msg != '=' &&	\
 	    *msg != ']' && *msg != '"')					\
 		++msg;							\
 	FAIL_IF("STRUCTURED-NAME", start == msg);			\
@@ -814,18 +898,18 @@ parsemsg_rfc5424(const char *from, int pri, char *msg)
 	IF_NOT_NILVALUE(buffer.sd) {
 		buffer.sd = msg;
 		/* SD-ELEMENT. */
-		while (*msg == '[') {
+		while (*msg && *msg == '[') {
 			++msg;
 			/* SD-ID. */
 			PARSE_SD_NAME();
 			/* SD-PARAM. */
-			while (*msg == ' ') {
+			while (*msg && *msg == ' ') {
 				++msg;
 				/* PARAM-NAME. */
 				PARSE_SD_NAME();
 				PARSE_CHAR("STRUCTURED-NAME", '=');
 				PARSE_CHAR("STRUCTURED-NAME", '"');
-				while (*msg != '"') {
+				while (*msg && *msg != '"') {
 					FAIL_IF("STRUCTURED-NAME",
 					    *msg == '\0');
 					if (*msg++ == '\\') {
@@ -952,7 +1036,7 @@ parsemsg_rfc3164(const char *from, int pri, char *msg)
 			 * This loop can only run for at most three
 			 * iterations before terminating.
 			 */
-			t_now = timer_now();
+			t_now = boot_time + timer_now();
 			localtime_r(&t_now, &tm_now);
 			for (year = tm_now.tm_year + 1;; --year) {
 				assert(year >= tm_now.tm_year - 1);
@@ -1066,12 +1150,15 @@ parsemsg(const char *from, char *msg)
 }
 
 /*
- * Take a raw input line from /dev/klog, split and format similar to syslog().
+ * Take a raw input line from /dev/klog, Linux /proc/klog, or Linux
+ * /dev/kmsg, split and format similar to syslog().
  */
 void printsys(char *msg)
 {
 	struct buf_msg buffer;
 	char line[MAXLINE + 1];
+	uint64_t ustime = 0;
+	uint64_t seqno = 0;
 	char *lp, *p, *q;
 	int c;
 
@@ -1084,11 +1171,54 @@ void printsys(char *msg)
 		buffer.msg = line;
 
 		if (*p == '<') {
+			/* /proc/klog or *BSD /dev/klog */
 			buffer.pri = 0;
 			while (isdigit(*++p))
 				buffer.pri = 10 * buffer.pri + (*p - '0');
 			if (*p == '>')
 				++p;
+		} else if (isdigit(*p)) {
+			/* Linux /dev/kmsg: "pri,seq#,msec,flag[,..];msg" */
+			time_t now = boot_time;
+
+			/* pri */
+			buffer.pri = 0;
+			while (isdigit(*p))
+				buffer.pri = 10 * buffer.pri + (*p++ - '0');
+
+			p++;	      /* skip ',' */
+
+			/* seq# */
+			while (isdigit(*p))
+				seqno = 10 * seqno + (*p++ - '0');
+
+			/*
+			 * Check if logged already (we've been restarted)
+			 * Account for wrap-around at 18446744073709551615
+			 */
+			if (sys_seqno > 0 && seqno <= sys_seqno) {
+				/* allow dupes around the edge */
+				if (sys_seqno < 18446744073709551000ULL)
+					return;
+			}
+			sys_seqno = seqno;
+
+			p++;	      /* skip ',' */
+
+			/* timestamp */
+			while (isdigit(*p))
+				ustime = 10 * ustime + (*p++ - '0');
+			now += ustime / 1000000;
+			buffer.timestamp.usec = ustime % 1000000;
+			localtime_r(&now, &buffer.timestamp.tm);
+
+			/* skip flags for now */
+			q = strchr(p, ';');
+			if (q)
+				p = ++q;
+		} else if (*p == ' ') {
+			/* Linux /dev/kmsg continuation line w/ SUBSYSTEM= DEVICE=, skip */
+			return;
 		} else {
 			/* kernel printf's come out on console */
 			buffer.flags |= IGN_CONS;
@@ -1098,8 +1228,18 @@ void printsys(char *msg)
 			buffer.pri = DEFSPRI;
 
 		q = lp;
-		while (*p != '\0' && (c = *p++) != '\n' && q < &line[MAXLINE])
+		while (*p != '\0' && (c = *p++) != '\n' && q < &line[MAXLINE]) {
+			/* Linux /dev/kmsg C-style hex encoding. */
+			if (c == '\\' && *p == 'x') {
+				char code[5] = "0x\0\0\0";
+
+				p++;
+				code[2] = *p++;
+				code[3] = *p++;
+				c = (int)strtol(code, NULL, 16);
+			}
 			*q++ = c;
+		}
 		*q = '\0';
 
 		logmsg(&buffer);
@@ -1126,10 +1266,11 @@ char *textpri(int pri)
 
 static void check_timestamp(struct buf_msg *buffer)
 {
-	struct logtime zero = { 0 };
+	struct logtime zero;
 	struct logtime now;
 	struct timeval tv;
 
+	memset(&zero, 0, sizeof(zero));
 	if (memcmp(&buffer->timestamp, &zero, sizeof(zero)))
 		return;
 
@@ -1693,7 +1834,7 @@ void wallmsg(struct filed *f, struct iovec *iov, int iovcnt)
 	 * and doing notty().
 	 */
 	if (fork() == 0) {
-		time_t t_now = timer_now();
+		time_t t_now = boot_time + timer_now();
 
 		(void)signal(SIGTERM, SIG_DFL);
 		(void)alarm(0);
@@ -2136,6 +2277,18 @@ static void signal_init(void)
 	SIGNAL(SIGCHLD, reapchild);
 }
 
+static void boot_time_init(void)
+{
+#ifdef __linux__
+	struct sysinfo si;
+	struct timeval tv;
+
+	gettimeofday(&tv, NULL);
+	sysinfo(&si);
+	boot_time = tv.tv_sec - si.uptime;
+#endif
+}
+
 /*
  *  INIT -- Initialize syslogd from configuration table
  */
@@ -2206,6 +2359,11 @@ static void init(void)
 		}
 	}
 
+	/*
+	 * Load / reload timezone data (in case it changed)
+	 */
+	tzset();
+
 	/*
 	 * Read configuration file(s)
 	 */
